import re
from semver import VersionInfo

from craft_ai.errors import CraftAiDecisionError
from craft_ai.time import Time
from craft_ai.timezones import get_timezone_key, timezone_offset_in_standard_format
from craft_ai.interpreter_v1 import InterpreterV1
from craft_ai.interpreter_v2 import InterpreterV2


class Interpreter(object):
    @staticmethod
    def decide(tree, args):
        bare_tree, configuration, tree_version = Interpreter._parse_tree(tree)
        interpreter = Interpreter._get_interpreter(tree_version)

        return Interpreter._decide(configuration, bare_tree, args, interpreter)

    ####################
    # Internal helpers #
    ####################

    @staticmethod
    def _decide(configuration, bare_tree, args, interpreter):
        if configuration != {}:
            time = None if len(args) == 1 else args[1]
            context_result = Interpreter._rebuild_context(configuration, args[0], time)
            context = context_result["context"]
        else:
            context = Interpreter.join_decide_args(args)
        # Convert timezones as integers into standard +/hh:mm format
        # This should only happen when no time generated value is required
        decide_context = Interpreter._convert_timezones_to_standard_format(
            configuration, context.copy()
        )

        decision = interpreter.decide(configuration, bare_tree, decide_context)
        decision["context"] = context

        return decision

    @staticmethod
    def _get_interpreter(tree_version):

        if tree_version >= VersionInfo(1, 0, 0) and tree_version < VersionInfo(2, 0, 0):
            return InterpreterV1
        elif tree_version >= VersionInfo(2, 0, 0) and tree_version < VersionInfo(
            3, 0, 0
        ):
            return InterpreterV2
        else:
            raise CraftAiDecisionError(
                """Invalid decision tree format, "{}" is currently not a valid version.""".format(
                    tree_version
                )
            )

    @staticmethod
    def _rebuild_context(configuration, state, time=None):

        missings = []
        # Model should come from _parse_tree and is assumed to be checked
        # upon already
        output = configuration["output"]
        context = configuration["context"]

        # We should not use the output key(s) to compare against
        configuration_ctx = {
            key: context[key] for (key, value) in context.items() if (key not in output)
        }

        # Check if we need the time object
        to_generate = []

        for prop in configuration_ctx.items():
            prop_name = prop[0]
            prop_attributes = prop[1]
            if prop_attributes["type"] in [
                "time_of_day",
                "day_of_week",
                "day_of_month",
                "month_of_year",
            ]:
                # is_generated is at True, we must generate the time for the
                # associated context property
                case_1 = (
                    "is_generated" in list(prop_attributes.keys())
                    and prop[1]["is_generated"]
                )
                # is_generated is not given, by default at True, so we must
                # generate it as well
                case_2 = "is_generated" not in list(prop_attributes.keys())
                if case_1 or case_2:
                    to_generate.append(prop_name)

        # Propagate missings properties to next function
        if to_generate:
            # Can't generate from time -> missings properties are errors
            if not isinstance(time, Time):
                # Check for missings properties
                for prop in to_generate:
                    if prop not in list(state.keys()):
                        missings.append(
                            "expected property '{}' is not defined".format(prop)
                        )

            # Generate context properties which need to
            else:
                for prop in to_generate:
                    state[prop] = time.to_dict()[configuration_ctx[prop]["type"]]

        # Rebuild the context with generated and non-generated values
        context = {
            feature: state.get(feature)
            for feature, properties in configuration_ctx.items()
            if feature in state
        }

        return {"context": context, "errors": missings}

    @staticmethod
    def join_decide_args(args):
        joined_args = {}
        for arg in args:
            if isinstance(arg, Time):
                joined_args.update(arg.to_dict())
            try:
                joined_args.update(arg)
            except TypeError:
                raise CraftAiDecisionError(
                    """Invalid context args, the given objects aren't dicts"""
                    """ or Time instances."""
                )
        return joined_args

    @staticmethod
    def _convert_timezones_to_standard_format(configuration, context):
        timezone_key = get_timezone_key(configuration["context"])
        if timezone_key and timezone_key in context:
            context[timezone_key] = timezone_offset_in_standard_format(
                context[timezone_key]
            )
        return context

    @staticmethod
    def _parse_tree(tree_object):
        # Checking definition of tree_object
        if not isinstance(tree_object, dict):
            raise CraftAiDecisionError(
                "Invalid decision tree format, the given json is not an object."
            )

        # Checking version existence
        tree_version = tree_object.get("_version")
        if not tree_version:
            raise CraftAiDecisionError(
                """Invalid decision tree format, unable to find the version"""
                """ informations."""
            )

        # Checking version and tree validity according to version
        if re.compile(r"\d+.\d+.\d+").match(tree_version) is None:
            raise CraftAiDecisionError(
                """Invalid decision tree format, "{}" is not a valid version.""".format(
                    tree_version
                )
            )
        elif tree_version >= VersionInfo(1, 0, 0) and tree_version < VersionInfo(
            3, 0, 0
        ):
            if tree_object.get("configuration") is None:
                raise CraftAiDecisionError(
                    """Invalid decision tree format, no configuration found"""
                )
            if tree_object.get("trees") is None:
                raise CraftAiDecisionError(
                    """Invalid decision tree format, no tree found."""
                )
            bare_tree = tree_object.get("trees")
            configuration = tree_object.get("configuration")
        else:
            raise CraftAiDecisionError(
                """Invalid decision tree format, {} is not a supported"""
                """ version.""".format(tree_version)
            )
        return bare_tree, configuration, tree_version



import math
import numbers

from craft_ai.errors import CraftAiDecisionError, CraftAiNullDecisionError
from craft_ai.operators import OPERATORS, OPERATORS_FUNCTION
from craft_ai.types import TYPES
from craft_ai.timezones import is_timezone

_DECISION_VERSION = "2.0.0"

_VALUE_VALIDATORS = {
    TYPES["continuous"]: lambda value: isinstance(value, numbers.Real),
    TYPES["enum"]: lambda value: isinstance(value, str),
    TYPES["boolean"]: lambda value: isinstance(value, bool),
    TYPES["timezone"]: lambda value: is_timezone(value),
    TYPES["time_of_day"]: lambda value: (
        isinstance(value, numbers.Real) and value >= 0 and value < 24
    ),
    TYPES["day_of_week"]: lambda value: (
        isinstance(value, int) and value >= 0 and value <= 6
    ),
    TYPES["day_of_month"]: lambda value: (
        isinstance(value, int) and value >= 1 and value <= 31
    ),
    TYPES["month_of_year"]: lambda value: (
        isinstance(value, int) and value >= 1 and value <= 12
    ),
}

############################
# Interpreter for V2 Trees #
############################


class InterpreterV2(object):
    @staticmethod
    def decide(configuration, bare_tree, context):
        InterpreterV2._check_context(configuration, context)

        decision_result = {}
        decision_result["output"] = {}
        for output in configuration.get("output"):
            output_type = configuration["context"][output]["type"]

            root = bare_tree[output]
            if not ("children" in root and len(root.get("children"))):
                # We check if a leaf has the key 'prediction' corresponging to a v2 tree
                prediction = root.get("prediction")
                if prediction is None:
                    prediction = root

                predicted_value = prediction.get("value")
                if predicted_value is None:
                    raise CraftAiNullDecisionError(
                        """Unable to take decision: the decision tree is not based"""
                        """ on any context operations."""
                    )

            decision_result["output"][output] = InterpreterV2._decide_recursion(
                bare_tree[output],
                context,
                bare_tree[output].get("output_values"),
                output_type,
                ["0"],
            )
        decision_result["_version"] = _DECISION_VERSION
        return decision_result

    # pylint: disable-msg=too-many-arguments, too-many-locals
    @staticmethod
    def _decide_recursion(node, context, output_values, output_type, path):
        # If we are on a leaf
        if not (node.get("children") is not None and len(node.get("children"))):
            # We check if a leaf has the key 'prediction' corresponging to a v2 tree
            prediction = node.get("prediction")
            if prediction is None:
                prediction = node

            predicted_value = prediction.get("value")
            if predicted_value is None:
                raise CraftAiNullDecisionError(
                    """Unable to take decision: the decision tree has no valid"""
                    """ predicted value for the given context.""",
                    {"decision_rules": [node.get("decision_rule")]},
                )

            leaf = {
                "predicted_value": predicted_value,
                "confidence": prediction.get("confidence") or 0,
                "decision_rules": [],
                "nb_samples": prediction["nb_samples"],
                "decision_path": "-".join(path),
            }

            distribution = prediction.get("distribution")
            if (
                not isinstance(distribution, list)
                and "standard_deviation" in distribution
            ):
                leaf["standard_deviation"] = distribution.get("standard_deviation")
                leaf["min"] = distribution.get("min")
                leaf["max"] = distribution.get("max")
            else:
                leaf["distribution"] = distribution

            return leaf
        # Finding the first element in this node's childrens matching the
        # operator condition with given context
        matching_child_i, matching_child = InterpreterV2._find_matching_child(
            node, context
        )

        # If there is no child corresponding matching the operators then we compute
        # the probabilistic distribution from this node.
        if not matching_child:
            return InterpreterV2.compute_distribution(
                node, output_values, output_type, path
            )
        # Add the matching child index to the path
        path.append(str(matching_child_i))
        # If a matching child is found, recurse
        try:
            result = InterpreterV2._decide_recursion(
                matching_child, context, output_values, output_type, path
            )
        except CraftAiDecisionError as err:
            metadata = err.metadata
            if node.get("decision_rule"):
                metadata["decision_rules"].insert(0, node["decision_rule"])
            raise CraftAiDecisionError(err.message, metadata)

        new_predicates = [
            {
                "property": matching_child["decision_rule"]["property"],
                "operator": matching_child["decision_rule"]["operator"],
                "operand": matching_child["decision_rule"]["operand"],
            }
        ]

        final_result = {
            "predicted_value": result["predicted_value"],
            "confidence": result["confidence"],
            "decision_rules": new_predicates + result["decision_rules"],
            "nb_samples": result["nb_samples"],
            "decision_path": result["decision_path"],
        }

        if result.get("standard_deviation", None) is not None:
            final_result["standard_deviation"] = result.get("standard_deviation")

        if result.get("min") is not None:
            final_result["min"] = result.get("min")

        if result.get("max") is not None:
            final_result["max"] = result.get("max")

        if result.get("distribution"):
            final_result["distribution"] = result.get("distribution")

        return final_result

    @staticmethod
    def compute_distribution(node, output_values, output_type, path):
        result = InterpreterV2._distribution(node, output_type)
        if output_type in ["enum", "boolean"]:
            distribution, nb_samples = result
            final_result = {
                "predicted_value": output_values[distribution.index(max(distribution))],
                "distribution": distribution,
                "nb_samples": nb_samples,
            }
        else:
            mean_value, nb_samples, standard_deviation = result
            final_result = {
                "predicted_value": mean_value,
                "nb_samples": nb_samples,
                "standard_deviation": standard_deviation,
            }
        final_result["decision_rules"] = []
        final_result["confidence"] = None
        final_result["decision_path"] = "-".join(path)

        return final_result

    @staticmethod
    def _distribution(node, output_type):
        # If it is a leaf
        if not (node.get("children") is not None and len(node.get("children"))):
            prediction = node["prediction"]
            value_distribution = prediction["distribution"]
            nb_samples = prediction["nb_samples"]
            if output_type in ["enum", "boolean"]:
                # It is a classification problem
                return [value_distribution, nb_samples]
            else:
                # It is a regression problem
                predicted_value = prediction.get("value")
                standard_deviation = value_distribution.get("standard_deviation")
                if predicted_value is not None:
                    if standard_deviation is not None:
                        return [predicted_value, nb_samples, standard_deviation]
                    raise CraftAiDecisionError(
                        """Unable to take decision: the decision tree has no valid"""
                        """ standard deviation for the given context."""
                    )
                raise CraftAiDecisionError(
                    """Unable to take decision: the decision tree has no valid"""
                    """ predicted value for the given context."""
                )

        # If it is not a leaf, we recurse into the children and store
        # the distributions/means and sizes of each child branch.
        def recurse(_child):
            return InterpreterV2._distribution(_child, output_type)

        values_sizes = list(map(recurse, node.get("children")))
        if output_type in ["enum", "boolean"]:
            # It is a classification problem
            values, sizes = zip(*values_sizes)
            return InterpreterV2.compute_mean_distributions(values, sizes)
        values, sizes, stds = zip(*values_sizes)
        return InterpreterV2.compute_mean_values(values, sizes, stds)

    @staticmethod
    def compute_mean_distributions(values, sizes):
        # Compute the weighted mean of the given array of distributions (array of probabilities).
        # Example, for values = [[ 4, 3, 6 ], [1, 2, 3], [3, 4, 5]], sizes = [1, 2, 1]
        # This function computes ([ 4, 3, 6]*1 + [1, 2, 3]*2 + [3, 4, 5]*6) / (1+2+1) = ...
        total_size = sum(sizes)
        ratio_applied = [
            [x * size / float(total_size) for x in x_array]
            for x_array, size in zip(values, sizes)
        ]
        return list(map(sum, zip(*ratio_applied))), total_size

    @staticmethod
    def compute_mean_values(values, sizes, stds=None):
        # Compute the weighted mean of the given array of values.
        # Example, for values = [ 4, 3, 6 ], sizes = [1, 2, 1]
        # This function computes (4*1 + 3*2 + 1*6) / (1+2+1) = 16/4 = 4
        # If no standard deviation array is given, use classical weighted mean formula:
        if stds is None:
            total_size = sum(sizes)
            mean = sum(
                [val * size / float(total_size) for val, size in zip(values, sizes)]
            )
            return mean, total_size
        # Otherwise, to compute the weighted standard deviation the following formula is used:
        # https://math.stackexchange.com/questions/2238086/calculate-variance-of-a-subset
        new_variance = None
        new_mean = None
        new_size = None
        for mean, std, size in zip(values, stds, sizes):
            variance = std * std
            if new_mean is None:
                new_variance = variance
                new_mean = mean
                new_size = size
                continue
            total_size = 1.0 * size + new_size
            if total_size == 0:
                continue
            new_variance = (1.0 / (total_size - 1.0)) * (
                (size - 1.0) * variance
                + (new_size - 1.0) * new_variance
                + (size * new_size / total_size) * (mean - new_mean) ** 2
            )
            new_mean = (1.0 / total_size) * (size * mean + new_size * new_mean)
            new_size = total_size
        return new_mean, new_size, math.sqrt(new_variance)

    @staticmethod
    def _find_matching_child(node, context):
        for child_index, child in enumerate(node["children"]):
            property_name = child["decision_rule"]["property"]
            operand = child["decision_rule"]["operand"]
            operator = child["decision_rule"]["operator"]
            context_value = context.get(property_name)

            if not isinstance(operator, str) or operator not in OPERATORS.values():
                raise CraftAiDecisionError(
                    """Invalid decision tree format, {} is not a valid"""
                    """ decision operator.""".format(operator)
                )

            if OPERATORS_FUNCTION[operator](context_value, operand):
                return child_index, child
        return None, {}

    @staticmethod
    def _check_context(configuration, context):
        # Extract the required properties (i.e. those that are not the output)
        expected_properties = [
            p for p in configuration["context"] if p not in configuration["output"]
        ]

        # Retrieve the missing properties
        missing_properties = [p for p in expected_properties if p not in context]

        # Validate the values
        bad_properties = [
            p
            for p in expected_properties
            if not InterpreterV2.validate_property_value(configuration, context, p)
        ]
        if missing_properties or bad_properties:
            missing_properties = sorted(missing_properties)
            missing_properties_messages = [
                "expected property '{}' is not defined".format(p)
                for p in missing_properties
            ]
            bad_properties = sorted(bad_properties)
            bad_properties_messages = [
                "'{}' is not a valid value for property '{}' of type '{}'".format(
                    context[p], p, configuration["context"][p]["type"]
                )
                for p in bad_properties
            ]

            errors = missing_properties_messages + bad_properties_messages

            # deal with missing properties
            if errors:
                message = (
                    "Unable to take decision, the given context is not valid: "
                    + errors.pop(0)
                )

                for error in errors:
                    message = "".join((message, ", ", error))
                message = message + "."

                metadata = {}
                if bad_properties:
                    metadata["badProperties"] = [
                        {
                            "property": p,
                            "type": configuration["context"][p]["type"],
                            "value": context[p],
                        }
                        for p in bad_properties
                    ]
                if missing_properties:
                    metadata["missingProperties"] = missing_properties

                raise CraftAiDecisionError(message, metadata)

    @staticmethod
    def validate_property_value(configuration, context, property_name):
        if context.get(property_name) is None:
            return True
        property_def = configuration["context"][property_name]
        property_type = property_def["type"]
        if property_type in _VALUE_VALIDATORS:
            property_value = context[property_name]
            return (
                _VALUE_VALIDATORS[property_type](property_value)
                or property_value is None
                or property_value == {}
            )
        return True



import re


def dict_depth(collection):
    if isinstance(collection, dict) and collection:
        return 1 + max(dict_depth(collection[a]) for a in collection)
    if isinstance(collection, list) and collection:
        return 1 + max(dict_depth(a) for a in collection)
    return 0


OPERATIONS_COUNT_RE = re.compile(r"\d+(?=\ operation\(s\))")


def extract_operations_count_from_message(message):
    count_matches = OPERATIONS_COUNT_RE.search(message)
    if not count_matches:
        return 0
    return int(count_matches.group())



import pandas as pd

from .. import Interpreter as VanillaInterpreter, Time
from ..errors import CraftAiNullDecisionError
from .utils import is_valid_property_value, create_timezone_df, format_input


class Interpreter(VanillaInterpreter):
    @staticmethod
    def decide_from_contexts_df(tree, contexts_df):
        bare_tree, configuration, tree_version = VanillaInterpreter._parse_tree(tree)
        interpreter = VanillaInterpreter._get_interpreter(tree_version)

        df = contexts_df.copy(deep=True)
        tz_col = [
            key
            for key, value in configuration["context"].items()
            if value["type"] == "timezone"
        ]
        # If a timezone is needed create a timezone dataframe which will
        # store the timezone to use. It can either be the DatetimeIndex
        # timezone or the timezone column if provided.
        if tz_col:
            tz_col = tz_col[0]
            df[tz_col] = create_timezone_df(contexts_df, tz_col).iloc[:, 0]

        predictions_iter = (
            Interpreter.decide_from_row(
                {
                    "bare_tree": bare_tree,
                    "context_ops": row,
                    "tz_col": tz_col,
                    "configuration": configuration,
                    "feature_names": df.columns.values,
                    "interpreter": interpreter,
                }
            )
            for row in df.itertuples(name=None)
        )
        return pd.DataFrame(predictions_iter, index=df.index)

    @staticmethod
    def decide_from_row(params):
        """
    Compute a decision from a valid context operation

    params : dict {
      "bare_tree": a valid craft ai tree,
      "context_row": a valid tuple representing a context operation,
      "tz_col": the time zone column,
      "configuration": a valid craft-ai configuration,
      "feature_names": the feature names,
      "interpreter": craft_ai interpreter
    }
    """

        context = {
            feature_name: format_input(value)
            for feature_name, value in zip(
                params["feature_names"], params["context_ops"][1:]
            )
            if is_valid_property_value(feature_name, value)
        }
        time = Time(
            t=params["context_ops"][0].value
            // 1000000000,  # Timestamp.value returns nanoseconds
            timezone=context[params["tz_col"]]
            if params["tz_col"]
            else params["context_ops"][0].tz,
        )
        try:
            decision = VanillaInterpreter._decide(
                params["configuration"],
                params["bare_tree"],
                (context, time),
                params["interpreter"],
            )

            return {
                "{}_{}".format(output, key): value
                for output, output_decision in decision["output"].items()
                for key, value in output_decision.items()
            }
        except CraftAiNullDecisionError as e:
            return {"error": e.message}



# To avoid conflicts between python's own 'time' and this 'time.py'
# cf. https://stackoverflow.com/a/28854227
from __future__ import absolute_import

import time

from datetime import datetime, tzinfo, timedelta

from pytz import utc as pyutc
from tzlocal import get_localzone
from dateutil.parser import isoparse

from craft_ai.errors import CraftAiTimeError
from craft_ai.timezones import is_timezone, timezone_offset_in_sec

_EPOCH = datetime(1970, 1, 1, tzinfo=pyutc)


class Time(object):
    """Handles time in a useful way for craft ai's client"""

    def __init__(self, t=None, timezone=None):
        def time_from_datetime_timestamp_and_timezone(timestamp, timezone):
            # Handle when datetime already provides timezone :
            # datetime(2012, 9, 12, 6, 0, 0, tzinfo=pytz.utc)
            result = timestamp
            if (result.tzinfo is None) and (not timezone):
                # Handle this format :
                # Time(datetime(2011, 1, 1, 0, 0), timezone=None)
                raise CraftAiTimeError("You must provide at least one timezone")
            elif (result.tzinfo is None) and timezone:
                # Handle this format :
                # Time(datetime(2011, 1, 1, 0, 0), timezone="+02:00")
                result = pyutc.localize(result)
                result = set_timezone(result, timezone)
            elif (result.tzinfo is not None) and (timezone):
                # Handle format like :
                # Time(datetime(2002, 10, 27, 6, 0, 0, tzinfo=utc),timezone="+02:00" )
                raise CraftAiTimeError(
                    "You must provide one timezone, but two were provided:"
                    " in the datetime and in the timezone parameter."
                )
            return result

        def time_from_string_timestamp_and_timezone(timestamp, timezone):
            # Else if t is a string we try to interprete it as an ISO time
            # string
            try:
                # Can't use strptime with %z in Python 2
                # https://stackoverflow.com/a/23940673
                result = isoparse(timestamp)
            except ValueError as e:
                raise CraftAiTimeError(
                    """Unable to instantiate Time from given string. {}""".format(
                        e.__str__()
                    )
                )

            if result.tzinfo is None:
                # Handle format like : Time(t="2017-01-01 00:00:00")
                if timezone:
                    # Handle format like : Time(t="2017-01-01 00:00:00", timezone="-03:00")
                    result = pyutc.localize(result)
                    result = set_timezone(result, timezone)
                else:
                    raise CraftAiTimeError(
                        "The given datetime string must be tz-aware,"
                        " or you must provide an explicit timezone."
                    )
            else:
                if timezone:
                    # Handle format like : Time("2011-04-22 01:00:00+0900", timezone="-03:00")
                    raise CraftAiTimeError(
                        "You must provide one timezone, but two were provided:"
                        " in the datetime string and in the timezone parameter."
                    )
            return result

        def set_time_and_timezone(timestamp, timezone):
            if timestamp is None:
                # If no initial timestamp is given, the current local time is used
                _time = datetime.now(get_localzone())
                # If a timezone is specified we can try to use it
                if timezone:
                    # Handle theses cases :   Time(timezone="+01:00") & Time(timezone="CET")
                    _time = set_timezone(_time, timezone)

            elif isinstance(timestamp, int):
                # Else if t is an int we try to use it as a given timestamp with
                # local UTC offset by default .
                try:
                    # Handle format like  : Time().timezone
                    _time = datetime.fromtimestamp(timestamp, get_localzone())
                except (OverflowError, OSError) as e:
                    raise CraftAiTimeError(
                        """Unable to instantiate Time from given timestamp. {}""".format(
                            e.__str__()
                        )
                    )
                # If a timezone is specified we can try to use it
                if timezone:
                    # Handle this type of datetime format : Time(1356998400, timezone="+0100")
                    _time = set_timezone(_time, timezone)

            elif isinstance(timestamp, datetime):
                _time = time_from_datetime_timestamp_and_timezone(timestamp, timezone)

            elif isinstance(timestamp, str):
                _time = time_from_string_timestamp_and_timezone(timestamp, timezone)

            else:
                raise CraftAiTimeError(
                    """Unable to instantiate Time from given timestamp."""
                    """ It must be integer or string."""
                )
            return _time

        def set_timezone(timestamp, timezone):
            if isinstance(timezone, tzinfo):
                # If it's already a timezone object, no more work is needed
                _time = timestamp.astimezone(timezone)
            elif is_timezone(timezone):
                # If it's a string, we convert it to a usable timezone object
                offset = timezone_offset_in_sec(timezone)
                _time = timestamp.astimezone(tz=dt_timezone(timedelta(seconds=offset)))
            else:
                raise CraftAiTimeError(
                    """Unable to instantiate Time with the given timezone."""
                    """ {} is neither a string nor a timezone.""".format(timezone)
                )
            return _time

        _time = set_time_and_timezone(t, timezone)

        try:
            self.utc_iso = _time.isoformat()
        except ValueError as e:
            raise CraftAiTimeError(
                """Unable to create ISO 8061 UTCstring. {}""".format(e.__str__())
            )

        self.day_of_week = _time.weekday()
        self.time_of_day = _time.hour + _time.minute / 60 + _time.second / 3600
        self.day_of_month = _time.day
        self.month_of_year = _time.month
        self.timezone = _time.strftime("%z")[:3] + ":" + _time.strftime("%z")[3:]
        self.timestamp = Time.timestamp_from_datetime(_time)

    def to_dict(self):
        """Returns the Time instance as a usable dictionary for craft_ai"""
        return {
            "timestamp": int(self.timestamp),
            "timezone": self.timezone,
            "time_of_day": self.time_of_day,
            "day_of_week": self.day_of_week,
            "day_of_month": self.day_of_month,
            "month_of_year": self.month_of_year,
            "utc_iso": self.utc_iso,
        }

    @staticmethod
    def timestamp_from_datetime(date_time):
        """Returns POSIX timestamp as float"""
        if date_time.tzinfo is None:
            return (
                time.mktime(
                    (
                        date_time.year,
                        date_time.month,
                        date_time.day,
                        date_time.hour,
                        date_time.minute,
                        date_time.second,
                        -1,
                        -1,
                        -1,
                    )
                )
                + date_time.microsecond / 1e6
            )
        return (date_time - _EPOCH).total_seconds()


# pylint: disable=C0103,W0212
class dt_timezone(tzinfo):
    """
    timezone class from python's standard library datetime
    This is reincluded here to ensure compatibility with python
    versions earlier than 3.2.
  """

    __slots__ = "_offset", "_name"

    # Sentinel value to disallow None
    _Omitted = object()

    def __new__(cls, offset, name=_Omitted):
        if not isinstance(offset, timedelta):
            raise TypeError("offset must be a timedelta")
        if name is cls._Omitted:
            if not offset:
                return cls.utc
            name = None
        elif not isinstance(name, str):
            raise TypeError("name must be a string")
        if not cls._minoffset <= offset <= cls._maxoffset:
            raise ValueError(
                "offset must be a timedelta"
                " strictly between -timedelta(hours=24) and"
                " timedelta(hours=24)."
            )
        if offset.microseconds != 0 or offset.seconds % 60 != 0:
            raise ValueError(
                "offset must be a timedelta" " representing a whole number of minutes"
            )
        return cls._create(offset, name)

    @classmethod
    def _create(cls, offset, name=None):
        self = tzinfo.__new__(cls)
        self._offset = offset
        self._name = name
        return self

    def __getinitargs__(self):
        """pickle support"""
        if self._name is None:
            return (self._offset,)
        return (self._offset, self._name)

    def __eq__(self, other):
        return self._offset == other._offset

    def __hash__(self):
        return hash(self._offset)

    def __repr__(self):
        """Convert to formal string, for repr().

    >>> tz = dt_timezone.utc
    >>> repr(tz)
    "datetime.dt_timezone.utc"
    >>> tz = dt_timezone(timedelta(hours=-5), 'EST')
    >>> repr(tz)
    "datetime.dt_timezone(datetime.timedelta(-1, 68400), 'EST')"
    """
        if self is self.utc:
            return "datetime.dt_timezone.utc"
        if self._name is None:
            return "%s(%r)" % ("datetime." + self.__class__.__name__, self._offset)
        return "%s(%r, %r)" % (
            "datetime." + self.__class__.__name__,
            self._offset,
            self._name,
        )

    def __str__(self):
        return self.tzname(None)

    def utcoffset(self, dt):
        if isinstance(dt, datetime) or dt is None:
            return self._offset
        raise TypeError("utcoffset() argument must be a datetime instance" " or None")

    def tzname(self, dt):
        if isinstance(dt, datetime) or dt is None:
            if self._name is None:
                return self._name_from_offset(self._offset)
            return self._name
        raise TypeError("tzname() argument must be a datetime instance" " or None")

    def dst(self, dt):
        if isinstance(dt, datetime) or dt is None:
            return None
        raise TypeError("dst() argument must be a datetime instance" " or None")

    def fromutc(self, dt):
        if isinstance(dt, datetime):
            if dt.tzinfo is not self:
                raise ValueError("fromutc: dt.tzinfo " "is not self")
            return dt + self._offset
        raise TypeError("fromutc() argument must be a datetime instance" " or None")

    _maxoffset = timedelta(hours=23, minutes=59)
    _minoffset = -_maxoffset

    @staticmethod
    def _name_from_offset(delta):
        if delta < timedelta(0):
            sign = "-"
            delta = -delta
        else:
            sign = "+"
        hours, rest = divmod(delta, timedelta(hours=1))
        minutes = rest // timedelta(minutes=1)
        return "UTC{}{:02d}:{:02d}".format(sign, hours, minutes)


dt_timezone.utc = dt_timezone._create(timedelta(0))
dt_timezone.min = dt_timezone._create(dt_timezone._minoffset)
dt_timezone.max = dt_timezone._create(dt_timezone._maxoffset)
# pylint: enable=C0103,W0212

# pylint: enable=import-self,ungrouped-imports,wrong-import-order,no-member



class CraftAiError(Exception):
    """Base class for exceptions in the craft ai client."""

    def __init__(self, message=None, metadata=None):
        self.message = message
        self.metadata = metadata
        super(CraftAiError, self).__init__(message, metadata)

    def __str__(self):
        return repr(self.message)


class CraftAiUnknownError(CraftAiError):
    """An unknown error occured in the craft ai client."""


class CraftAiNetworkError(CraftAiError):
    """A network error occured between the client and craft ai."""


class CraftAiCredentialsError(CraftAiError):
    """A credentials error occured."""


class CraftAiInternalError(CraftAiError):
    """An Internal Server Error (500) ocurred on craft ai's side."""


class CraftAiBadRequestError(CraftAiError):
    """An unvalid request was send to craft ai's API."""


class CraftAiNotFoundError(CraftAiError):
    """A Not Found Error (404) occured on craft ai's side."""


class CraftAiDecisionError(CraftAiError):
    """An error occured during the decision phase."""


class CraftAiNullDecisionError(CraftAiDecisionError):
    """An error occured during the decision phase."""


class CraftAiTimeError(CraftAiError):
    """An error occured during the creation of a craft_ai.Time instance."""


class CraftAiTokenError(CraftAiError):
    """An invalid token error occured."""


class CraftAiLongRequestTimeOutError(CraftAiError):
    """Request timed out because the computation is not finished, please try again."""



import json
import pandas as pd

from .. import Client as VanillaClient
from ..constants import DEFAULT_DECISION_TREE_VERSION
from ..errors import CraftAiBadRequestError
from .interpreter import Interpreter
from .utils import format_input, is_valid_property_value, create_timezone_df


def chunker(to_be_chunked_df, chunk_size):
    return (
        to_be_chunked_df[pos : pos + chunk_size]
        for pos in range(0, len(to_be_chunked_df), chunk_size)
    )


class Client(VanillaClient):
    """Client class for craft ai's API using pandas dataframe types"""

    def add_agent_operations(self, agent_id, operations):
        if isinstance(operations, pd.DataFrame):
            if not isinstance(operations.index, pd.DatetimeIndex):
                raise CraftAiBadRequestError(
                    "Invalid dataframe given, it is not time indexed."
                )
            if operations.index.tz is None:
                raise CraftAiBadRequestError(
                    """tz-naive DatetimeIndex are not supported,
                                     it must be tz-aware."""
                )
            agent = super(Client, self).get_agent(agent_id)
            operations = operations.copy(deep=True)

            tz_col = [
                key
                for key, value in agent["configuration"]["context"].items()
                if value["type"] == "timezone"
            ]
            if tz_col:
                tz_col = tz_col[0]
                operations[tz_col] = create_timezone_df(operations, tz_col).iloc[:, 0]

            chunk_size = self.config["operationsChunksSize"]
            for chunk in chunker(operations, chunk_size):
                chunk_operations = [
                    {
                        "timestamp": row.name.value
                        // 10 ** 9,  # Timestamp.value returns nanoseconds
                        "context": {
                            col: format_input(row[col])
                            for col in chunk.columns
                            if is_valid_property_value(col, row[col])
                        },
                    }
                    for _, row in chunk.iterrows()
                ]
                super(Client, self).add_agent_operations(agent_id, chunk_operations)

            return {
                "message": 'Successfully added %i operation(s) to the agent "%s/%s/%s" context.'
                % (
                    len(operations),
                    self.config["owner"],
                    self.config["project"],
                    agent_id,
                )
            }
        else:
            return super(Client, self).add_agent_operations(agent_id, operations)

    def add_agents_operations_bulk(self, payload):
        """Add operations to a group of agents.

        :param list payload: contains the informations necessary for the action.
        It's in the form [{"id": agent_id, "operations": operations}]
        With id that is an str containing only characters in "a-zA-Z0-9_-"
        and must be between 1 and 36 characters. It must referenced an
        existing agent.
        With operations either a list of dict or a DataFrame that has
        the form given in the craft_ai documentation and the configuration of
        the agent.

        :return: list of agents containing a message about the added
        operations.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if all of the ids are invalid or
        referenced non existing agents or one of the operations is invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, _, _ = self._check_entity_id_bulk(
            payload, check_serializable=False
        )
        valid_payload = [payload[i] for i in valid_indices]

        new_payload = []
        for agent in valid_payload:
            operations = agent["operations"]
            if isinstance(operations, pd.DataFrame):
                if not isinstance(operations.index, pd.DatetimeIndex):
                    raise CraftAiBadRequestError(
                        "Invalid dataframe given for agent "
                        "{}, it is not time indexed.".format(agent["id"])
                    )
                if operations.index.tz is None:
                    raise CraftAiBadRequestError(
                        "tz-naive DatetimeIndex are not supported for "
                        "agent {}, it must be tz-aware.".format(agent["id"])
                    )

                new_operations = [
                    {
                        "timestamp": row.name.value
                        // 10 ** 9,  # Timestamp.value returns nanoseconds
                        "context": {
                            col: row[col]
                            for col in operations.columns
                            if is_valid_property_value(col, row[col])
                        },
                    }
                    for _, row in operations.iterrows()
                ]
                new_payload.append({"id": agent["id"], "operations": new_operations})
            elif isinstance(operations, list):
                # Check if the operations are serializable
                json.dumps([agent])
                new_payload.append({"id": agent["id"], "operations": operations})
            else:
                raise CraftAiBadRequestError(
                    "The operations are not put in a DataFrame or a list"
                    "of dict form for the agent {}.".format(agent["id"])
                )

        return super(Client, self).add_agents_operations_bulk(new_payload)

    def get_agent_operations(self, agent_id, start=None, end=None):
        operations_list = super(Client, self).get_agent_operations(agent_id, start, end)
        return pd.DataFrame(
            [operation["context"] for operation in operations_list],
            index=pd.to_datetime(
                [operation["timestamp"] for operation in operations_list], unit="s"
            ).tz_localize("UTC"),
        )

    def get_agent_states(self, agent_id, start=None, end=None):
        states = super(Client, self).get_agent_states(agent_id, start, end)

        return pd.DataFrame(
            [state["sample"] for state in states],
            index=pd.to_datetime(
                [state["timestamp"] for state in states], unit="s"
            ).tz_localize("UTC"),
        )

    @staticmethod
    def decide_from_contexts_df(tree, contexts_df):
        if isinstance(contexts_df, pd.DataFrame):
            if not isinstance(contexts_df.index, pd.DatetimeIndex):
                raise CraftAiBadRequestError(
                    "Invalid dataframe given, it is not time indexed."
                )
            if contexts_df.index.tz is None:
                raise CraftAiBadRequestError(
                    """tz-naive DatetimeIndex are not supported,
                                     it must be tz-aware."""
                )
        else:
            raise CraftAiBadRequestError("Invalid data given, it is not a DataFrame.")
        return Interpreter.decide_from_contexts_df(tree, contexts_df)

    def get_agent_decision_tree(
        self, agent_id, timestamp=None, version=DEFAULT_DECISION_TREE_VERSION
    ):
        # Convert pandas timestamp to a numerical timestamp in seconds
        if isinstance(timestamp, pd.Timestamp):
            timestamp = timestamp.value // 10 ** 9

        return super(Client, self).get_agent_decision_tree(agent_id, timestamp, version)

    def get_generator_decision_tree(
        self, generator_id, timestamp=None, version=DEFAULT_DECISION_TREE_VERSION
    ):
        # Convert pandas timestamp to a numerical timestamp in seconds
        if isinstance(timestamp, pd.Timestamp):
            timestamp = timestamp.value // 10 ** 9

        return super(Client, self).get_generator_decision_tree(
            generator_id, timestamp, version
        )



from copy import copy
from .errors import CraftAiError


def _update_paths(paths, idx):
    """ add new path build on idx to all paths """
    if paths:
        return paths + ["{}-{}".format(paths[-1], idx)]
    return [str(idx)]


def _paths(tree, paths=None):
    """ return a raw list of all paths in a tree """
    if paths is None:
        paths = ["0"]
    if "children" in tree:
        current_paths = copy(paths)
        for i, child in enumerate(tree["children"]):
            paths.extend(_paths(child, _update_paths(current_paths, i)))
        return paths
    return paths


def _get_paths(tree):
    """ return a set of all paths in a tree """
    return set(_paths(tree))


def _is_neighbour(path0, path1):
    """
        Boolean function. A neighbour has exactly the same path excepted for the last node
    """
    return path0[:-1] == path1[:-1] and path0 != path1


def _get_neighbours(paths, decision_path):
    """
    Collect all neighbours paths of the given decision path
    param: paths: paths aggregator
    param: decision_path: decision path to get neighbours from
    """
    split = decision_path.split("-")
    neighbours = []
    for step in range(1, len(split) + 1):
        for path in paths:
            if _is_neighbour(path, "-".join(split[:step])):
                neighbours.append(path)
    return neighbours


def extract_output_tree(tree, output_property=None):
    """
    Extract the output decision tree specific for a given output property from a full decision tree.

    This function accepts trees as retrieved from `craft_ai.Client.get_generator_decision_tree`.

    Parameters:
        tree: A tree.
        output_property (optional): If provided, the output property for which the tree predicts
            values, otherwise the first defined tree is retrieved.
    """
    if not isinstance(tree, dict):
        raise CraftAiError(
            """Unable to retrieve the output tree, """
            """the given decision tree format is not a valid, expected a 'dict' got a {}.""".format(
                type(tree)
            )
        )
    if "trees" not in tree:
        raise CraftAiError(
            """Unable to retrieve the output tree, """
            """the given decision tree format is not a valid, no 'trees' property defined."""
        )
    trees = tree["trees"]
    if not output_property:
        output_property = list(trees)[0]
    if output_property not in trees:
        raise CraftAiError(
            """'{}' output tree can't be found in the given decision tree.""".format(
                output_property
            )
        )
        tree = tree["trees"][output_property]
    return trees[output_property]


def extract_decision_paths_from_tree(tree):
    """
    Retrieve all the decision paths from a tree.

    This function accepts trees as retrieved from `craft_ai.Client.get_generator_decision_tree`.

    Parameters:
        tree: A tree.
    Returns: e.g. ['0', '0-0', '0-1']
    """
    return _get_paths(extract_output_tree(tree))


def extract_decision_path_neighbors(
    tree, decision_path, max_depth=None, include_self=False
):
    """
    Retrieve neighbor of a decision path in a tree.

    This function accepts trees as retrieved from `craft_ai.Client.get_generator_decision_tree`.

    Parameters:
        tree: A tree.
        decision_path: string tree path eg. "0-2-1".
        max_depth (int, optional): positive int filter neighbours on their depth,
            default is None.
        include_self (bool, optional): include the given decision_path to the neighbours,
            default is False.
    """
    paths = _get_paths(extract_output_tree(tree))
    if decision_path not in paths:
        raise CraftAiError(
            """Invalid decision path given. """
            """{} not found in tree""".format(decision_path)
        )

    dp_depth = len(decision_path.split("-"))
    neighbours = _get_neighbours(paths, decision_path)
    if max_depth is None:
        max_depth = dp_depth
    if max_depth < 0:
        raise CraftAiError(
            """Invalid max depth given: {} should be None or a positive integer """.format(
                max_depth
            )
        )
    filtered = [n for n in neighbours if len(n.split("-")) <= max_depth]
    if include_self:
        filtered.append(decision_path)
    return filtered



from functools import reduce as ft_reduce

from .errors import CraftAiError
from .formatters import format_decision_rules
from .operators import OPERATORS


def _is_is_reducer(rule_1, rule_2):
    if rule_1["operand"] and (rule_1["operand"] != rule_2["operand"]):
        raise CraftAiError(
            "Operator '{}' can't have different value. Set to '{}' and receive '{}'".format(
                OPERATORS["IS"], rule_1["operand"], rule_2["operand"]
            )
        )
    return {
        "property": rule_1["property"],
        "operator": OPERATORS["IS"],
        "operand": rule_2["operand"],
    }


def _in_in_reducer(rule_1, rule_2):
    op_1_from = rule_1["operand"][0]
    op_1_to = rule_1["operand"][1]

    op_2_from = rule_2["operand"][0]
    op_2_to = rule_2["operand"][1]

    op_1_is_cyclic = op_1_from > op_1_to
    op_2_is_cyclic = op_2_from > op_2_to
    op_2_from_in_op_1 = (
        (op_2_from >= op_1_from or op_2_from <= op_1_to)
        if op_1_is_cyclic
        else (op_2_from >= op_1_from and op_2_from <= op_1_to)
    )
    op_2_to_in_op_1 = (
        (op_2_to >= op_1_from or op_2_to <= op_1_to)
        if op_1_is_cyclic
        else (op_2_to >= op_1_from and op_2_to <= op_1_to)
    )
    op_1_from_in_op_2 = (
        (op_1_from >= op_2_from or op_1_from <= op_2_to)
        if op_2_is_cyclic
        else (op_1_from >= op_2_from and op_1_from <= op_2_to)
    )
    op_1_to_in_op_2 = (
        (op_1_to >= op_2_from or op_1_to <= op_2_to)
        if op_2_is_cyclic
        else (op_1_to >= op_2_from and op_1_to <= op_2_to)
    )

    if op_1_from_in_op_2 and op_1_to_in_op_2:
        # op_1 belongs to op_2
        #    |    op_1    |
        #   |       op_2       |
        return rule_1

    if op_2_from_in_op_1 and op_2_to_in_op_1:
        # op_2 belongs to op_1
        #   |    op_1    |
        #      |  op_2  |
        return rule_2

    if op_2_from_in_op_1 and op_1_to_in_op_2:
        # overlap 1
        #       |    op_1    |
        #           |   op_2   |
        return {
            "property": rule_1["property"],
            "operator": OPERATORS["IN_INTERVAL"],
            "operand": [op_2_from, op_1_to],
        }

    if op_2_to_in_op_1 and op_1_from_in_op_2:
        # overlap 2
        #        |    op_1    |
        #     |   op_2   |
        return {
            "property": rule_1["property"],
            "operator": OPERATORS["IN_INTERVAL"],
            "operand": [op_1_from, op_2_to],
        }

    # disjointed
    #    |    op_1    |
    #                  |   op_2   |
    raise CraftAiError(
        """Unable to reduce decision rules '{}' and '{}': """
        """the resulting rule is not fulfillable.""".format(
            format_decision_rules([rule_1]), format_decision_rules([rule_2])
        )
    )


def _in_gte_reducer(rule_1, rule_2):
    op_1_from = rule_1["operand"][0]
    op_1_to = rule_1["operand"][1]
    op_2 = rule_2["operand"]

    op_1_is_cyclic = op_1_from > op_1_to

    if op_1_is_cyclic:
        # Cyclics makes no sense with single bound limits
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """the resulting rule is not fulfillable.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )

    if op_2 >= op_1_to:
        # op_2 after op_1, disjointed
        #    |    op_1    |
        #                  |op_2
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """the resulting rule is not fulfillable.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )

    if op_2 >= op_1_from and op_2 < op_1_to:
        # op_2 belongs to op_1
        #    |    op_1    |
        #           |op_2
        return {
            "property": rule_1["property"],
            "operator": OPERATORS["IN_INTERVAL"],
            "operand": [op_2, op_1_to],
        }

    # op_2 before op_1
    #    |    op_1    |
    #   |op_2
    return rule_1


def _in_lt_reducer(rule_1, rule_2):
    op_1_from = rule_1["operand"][0]
    op_1_to = rule_1["operand"][1]
    op_2 = rule_2["operand"]

    op_1_is_cyclic = op_1_from > op_1_to

    if op_1_is_cyclic:
        # Cyclics makes no sense with single bound limits
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """the resulting rule is not fulfillable.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )

    if op_2 < op_1_from:
        # op_2 before op_1, disjointed
        #      |    op_1    |
        # op_2|
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """the resulting rule is not fulfillable.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )

    if op_2 >= op_1_from and op_2 < op_1_to:
        # op_2 belongs to op_1
        #    |    op_1    |
        #           |op_2
        return {
            "property": rule_1["property"],
            "operator": OPERATORS["IN_INTERVAL"],
            "operand": [op_1_from, op_2],
        }

    # op_2 after op_1
    #    |    op_1    |
    #                 op_2|
    return rule_1


def _gte_lt_reducer(rule_1, rule_2):
    new_lower_bound = rule_1["operand"]
    new_upper_bound = rule_2["operand"]
    if new_upper_bound < new_lower_bound:
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """the resulting rule is not fulfillable.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )
    return {
        "property": rule_1["property"],
        "operator": OPERATORS["IN_INTERVAL"],
        "operand": [new_lower_bound, new_upper_bound],
    }


REDUCER_FROM_DECISION_RULE = {
    OPERATORS["IS"]: {OPERATORS["IS"]: _is_is_reducer},
    OPERATORS["IN_INTERVAL"]: {
        OPERATORS["IN_INTERVAL"]: _in_in_reducer,
        OPERATORS["GTE"]: _in_gte_reducer,
        OPERATORS["LT"]: _in_lt_reducer,
    },
    OPERATORS["GTE"]: {
        OPERATORS["IN_INTERVAL"]: lambda rule_1, rule_2: _in_gte_reducer(
            rule_2, rule_1
        ),
        OPERATORS["GTE"]: lambda rule_1, rule_2: {
            "property": rule_1["property"],
            "operator": OPERATORS["GTE"],
            "operand": max(rule_1["operand"], rule_2["operand"]),
        },
        OPERATORS["LT"]: _gte_lt_reducer,
    },
    OPERATORS["LT"]: {
        OPERATORS["IN_INTERVAL"]: lambda rule_1, rule_2: _in_lt_reducer(rule_2, rule_1),
        OPERATORS["GTE"]: lambda rule_1, rule_2: _gte_lt_reducer(rule_2, rule_1),
        OPERATORS["LT"]: lambda rule_1, rule_2: {
            "property": rule_1["property"],
            "operator": OPERATORS["LT"],
            "operand": min(rule_1["operand"], rule_2["operand"]),
        },
    },
}


def _decision_rules_reducer(rule_1, rule_2):
    if rule_1 is None or rule_2 is None:
        return rule_1 if rule_1 is not None else rule_2
    if (
        rule_1["operator"] not in REDUCER_FROM_DECISION_RULE
        or rule_2["operator"] not in REDUCER_FROM_DECISION_RULE[rule_1["operator"]]
    ):
        raise CraftAiError(
            """Unable to reduce decision rules '{}' and '{}': """
            """incompatible operators.""".format(
                format_decision_rules([rule_1]), format_decision_rules([rule_2])
            )
        )
    return REDUCER_FROM_DECISION_RULE[rule_1["operator"]][rule_2["operator"]](
        rule_1, rule_2
    )


def _unique_seq(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def reduce_decision_rules(rules):
    properties = _unique_seq([rule["property"] for rule in rules])
    return [
        ft_reduce(
            _decision_rules_reducer, [rule for rule in rules if rule["property"] == p]
        )
        for p in properties
    ]



# To avoid conflicts between python's own 'time' and this 'time.py'
# cf. https://stackoverflow.com/a/28854227
from __future__ import absolute_import

import json
import time
import datetime

from platform import python_implementation, python_version
from urllib.parse import urlparse

import requests

from . import __version__ as pkg_version
from .constants import AGENT_ID_PATTERN, DEFAULT_DECISION_TREE_VERSION
from .errors import (
    CraftAiCredentialsError,
    CraftAiBadRequestError,
    CraftAiNotFoundError,
    CraftAiUnknownError,
    CraftAiInternalError,
    CraftAiLongRequestTimeOutError,
    CraftAiNetworkError,
)
from .helpers import extract_operations_count_from_message
from .interpreter import Interpreter
from .jwt_decode import jwt_decode

USER_AGENT = "craft-ai-client-python/{} [{} {}]".format(
    pkg_version, python_implementation(), python_version()
)

ERROR_ID_MESSAGE = (
    "Invalid agent id given. "
    "It must be a string containing only "
    'characters in "a-zA-Z0-9_-" '
    "and must be between 1 and 36 characters."
)


def current_time_ms():
    return int(round(time.time() * 1000))


class Client(object):
    """Client class for craft ai's API"""

    def __init__(self, cfg):
        self._base_url = ""
        self._headers = {}
        self._config = {}
        # Requests session: connection pooling and base configuration for all requests
        self._requests_session = requests.Session()

        try:
            self.config = cfg
        except (CraftAiCredentialsError, CraftAiBadRequestError) as err:
            raise err

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, cfg):
        cfg = cfg.copy()
        (payload, _, _, _) = jwt_decode(cfg.get("token"))
        cfg["owner"] = cfg["owner"] if "owner" in cfg else payload.get("owner")
        cfg["project"] = cfg["project"] if "project" in cfg else payload.get("project")
        cfg["url"] = cfg["url"] if "url" in cfg else payload.get("platform")

        if not isinstance(cfg.get("project"), str):
            raise CraftAiCredentialsError(
                """Unable to create client with no"""
                """ or invalid project provided."""
            )
        else:
            splitted_project = cfg.get("project").split("/")
            if len(splitted_project) == 2:
                cfg["owner"] = splitted_project[0]
                cfg["project"] = splitted_project[1]
            elif len(splitted_project) > 2:
                raise CraftAiCredentialsError(
                    """Unable to create client with invalid""" """ project name."""
                )
        if not isinstance(cfg.get("owner"), str):
            raise CraftAiCredentialsError(
                """Unable to create client with no""" """ or invalid owner provided."""
            )
        if not isinstance(cfg.get("operationsChunksSize"), int):
            cfg["operationsChunksSize"] = 200
        if cfg.get("decisionTreeRetrievalTimeout") is not False and not isinstance(
            cfg.get("decisionTreeRetrievalTimeout"), int
        ):
            cfg["decisionTreeRetrievalTimeout"] = 1000 * 60 * 5  # 5 minutes
        if not isinstance(cfg.get("url"), str):
            cfg["url"] = "https://beta.craft.ai"
        if cfg.get("url").endswith("/"):
            raise CraftAiBadRequestError(
                """Unable to create client with"""
                """ invalid url provided. The url"""
                """ should not terminate with a"""
                """ slash."""
            )
        self._config = cfg

        self._base_url = "{}/api/v1/{}/{}".format(
            self.config["url"], self.config["owner"], self.config["project"]
        )

        if cfg.get("proxy"):
            scheme = urlparse(self.config["url"]).scheme
            if not scheme:
                raise CraftAiCredentialsError(
                    """Unable to create client with an URL"""
                    """ without a scheme. Cannot configure"""
                    """ the proxy."""
                )
            proxies = {}
            proxies[scheme] = cfg.get("proxy")
            self._requests_session.proxies = proxies
        # Headers have to be set here to avoid multiple definitions
        # of the 'Authorization' header if config is modified
        base_headers = {}
        base_headers["Authorization"] = "Bearer " + self.config.get("token")
        base_headers["User-Agent"] = USER_AGENT
        self._requests_session.headers = base_headers

    #################
    # Agent methods #
    #################

    def create_agent(self, configuration, agent_id=""):
        """Create an agent.

        :param dict configuration: Form given by the craft_ai documentation.
        :param str agent_id: Optional. The id of the agent to create. It
        must be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.
        :default agent_id: "", the agent_id is generated.

        :return: agent created.
        :rtype: dict.

        :raise CraftAiBadRequestError: if the input is not of
        the right form.
        """
        # Extra header in addition to the main session's
        ct_header = {"Content-Type": "application/json; charset=utf-8"}

        # Building payload and checking that it is valid for a JSON
        # serialization
        payload = {"configuration": configuration}

        if agent_id != "":
            # Raises an error when agent_id is invalid
            self._check_entity_id(agent_id)

            payload["id"] = agent_id

        try:
            json_pl = json.dumps(payload)
        except TypeError as err:
            raise CraftAiBadRequestError(
                "Invalid configuration or agent id given. {}".format(err.__str__())
            )

        req_url = "{}/agents".format(self._base_url)
        resp = self._requests_session.post(req_url, headers=ct_header, data=json_pl)

        agent = self._decode_response(resp)

        return agent

    def create_agents_bulk(self, payload):
        """Create a group of agents.

        :param list payload: Contains the informations to create the agents.
        It's in the form [{"id": agent_id, "configuration": configuration}]
        With an optional id key that is an str containing only characters
        in "a-zA-Z0-9_-" and must be between 1 and 36 characters.
        With configuration having the form given in the craft_ai documentation.

        :return: agents created which are represented with dictionnaries.
        :rtype: List of dict.

        :raises CraftAiBadRequestError: If all of the ids or all of the
        configurations are invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_agents = self._check_entity_id_bulk(
            payload
        )

        # Create the json file with the agents with valid id and send it
        valid_agents = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/agents".format(self._base_url),
            "POST",
        )

        if invalid_indices == []:
            return valid_agents

        # Put the valid and invalid agents in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_agents, invalid_indices, invalid_agents
        )

    def get_agent(self, agent_id):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}".format(self._base_url, agent_id)
        resp = self._requests_session.get(req_url)

        agent = self._decode_response(resp)

        return agent

    def list_agents(self):

        req_url = "{}/agents".format(self._base_url)
        resp = self._requests_session.get(req_url)

        agents = self._decode_response(resp)

        return agents["agentsList"]

    def delete_agent(self, agent_id):
        """Delete an agent.

        :param str agent_id: The id of the agent to delete. It must
        be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.

        :return: agent deleted.
        :rtype: dict.
        """
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}".format(self._base_url, agent_id)
        resp = self._requests_session.delete(req_url)

        decoded_resp = self._decode_response(resp)

        return decoded_resp

    def delete_agents_bulk(self, payload):
        """Delete a group of agents

        :param list payload: Contains the informations to delete the agents.
        It's in the form [{"id": agent_id}].
        With id an str containing only characters in "a-zA-Z0-9_-" and must
        be between 1 and 36 characters.

        :return: the list of agents deleted which are represented with
        dictionnaries.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: If all of the ids are invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_agents = self._check_entity_id_bulk(
            payload
        )

        # Create the json file with the agents with valid id and send it
        valid_agents = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/agents".format(self._base_url),
            "DELETE",
        )

        if invalid_indices == []:
            return valid_agents

        # Put the valid and invalid agents in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_agents, invalid_indices, invalid_agents
        )

    def get_shared_agent_inspector_url(self, agent_id, timestamp=None):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}/shared".format(self._base_url, agent_id)
        resp = self._requests_session.get(req_url)

        url = self._decode_response(resp)

        if timestamp is not None:
            return "{}?t={}".format(url["shortUrl"], str(timestamp))

        return url["shortUrl"]

    def delete_shared_agent_inspector_url(self, agent_id):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}/shared".format(self._base_url, agent_id)
        resp = self._requests_session.delete(req_url)

        decoded_resp = self._decode_response(resp)

        return decoded_resp

    ####################
    # Generator method #
    ####################

    def create_generator(self, configuration, generator_id=""):
        """ Create a generator.

        :param dict configuration: Form given by the craft_ai documentation.
        :param str generator_id: The id of the generator to create. It must be
        an str containing only characters in "a-zA-Z0-9_-" and must be
        between 1 and 36 characters.
        :param default generator_id : "" In this case the generator_id is
        generated.
        """
        # Extra header in addition to the main session's
        ct_header = {"Content-Type": "application/json; charset=utf-8"}

        # Building payload and checking that it is valid for a JSON
        # serialization

        payload = {"configuration": configuration}

        if generator_id != "":
            # Raises an error when generator_id is invalid
            self._check_entity_id(generator_id)

            payload["id"] = generator_id

        try:
            json_pl = json.dumps(payload)
        except TypeError as err:
            raise CraftAiBadRequestError(
                "Invalid configuration or generator id given. {}".format(err.__str__())
            )

        req_url = "{}/generators".format(self._base_url)
        resp = self._requests_session.post(req_url, headers=ct_header, data=json_pl)

        generator = self._decode_response(resp)

        return generator

    def create_generators_bulk(self, payload):
        """Create a group of generators.

        :param list payload: Contains the informations to create the generators.
        It's in the form [{"id": generator_id, "configuration": configuration}]
        With an id key that is an str containing only characters
        in "a-zA-Z0-9_-" and must be between 1 and 36 characters.
        With configuration having the form given in the craft_ai documentation.

        :return: Generators created which are represented with dictionnaries.
        :rtype: List of dict.

        :raises CraftAiBadRequestError: If all of the ids or all of the
        configurations are invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_generators = self._check_entity_id_bulk(
            payload
        )
        # Create the json file with the generators with valid id and send it
        valid_generators = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/generators".format(self._base_url),
            "POST",
        )

        if invalid_indices == []:
            return valid_generators
        # Put the valid and invalid generators in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_generators, invalid_indices, invalid_generators
        )

    def get_generator(self, generator_id):
        # Raises an error when generator_id is invalid
        self._check_entity_id(generator_id)

        req_url = "{}/generators/{}".format(self._base_url, generator_id)
        resp = self._requests_session.get(req_url)

        generator = self._decode_response(resp)

        return generator

    def list_generators(self):

        req_url = "{}/generators".format(self._base_url)
        resp = self._requests_session.get(req_url)

        generators = self._decode_response(resp)

        return generators["generatorsList"]

    def delete_generator(self, generator_id):
        """ Delete a generator

        :param str generator_id: The id of the generator to delete. It must be
        an str containing only characters in "a-zA-Z0-9_-" and must be
        between 1 and 36 characters. It must reference an existing generator.
        """

        # Raises an error when generator_id is invalid
        self._check_entity_id(generator_id)

        req_url = "{}/generators/{}".format(self._base_url, generator_id)
        resp = self._requests_session.delete(req_url)

        decoded_resp = self._decode_response(resp)

        return decoded_resp

    def delete_generators_bulk(self, payload):
        """Delete a group of generators

        :param list payload: Contains the informations to delete the generators.
        It's in the form [{"id": generator_id}].
        With id an str containing only characters in "a-zA-Z0-9_-" and must
        be between 1 and 36 characters.

        :return: the list of generators deleted which are represented with
        dictionnaries.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: If all of the ids are invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_generators = self._check_entity_id_bulk(
            payload
        )

        # Create the json file with the generators with valid id and send it
        valid_generators = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/generators".format(self._base_url),
            "DELETE",
        )

        if invalid_indices == []:
            return valid_generators

        # Put the valid and invalid generators in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_generators, invalid_indices, invalid_generators
        )

    def _get_generator_decision_tree(
        self, generator_id, timestamp, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Tool for the function get_generator_decision_tree.

        :param str generator_id: the id of the generator whose tree to get. It
        must be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.
        :param int timestamp: Optional. The decision tree is comptuted
        at this timestamp.
        :default timestamp: None, means that we get the tree computed
        with all its context history.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: decision tree.
        :rtype: dict.
        """
        self._requests_session.headers["x-craft-ai-tree-version"] = version

        # If we give no timestamp the default behaviour is to give the tree
        # from the latest timestamp
        if timestamp is None:
            req_url = "{}/generators/{}/tree?".format(self._base_url, generator_id)
        else:
            req_url = "{}/generators/{}/tree?t={}".format(
                self._base_url, generator_id, timestamp
            )

        resp = self._requests_session.get(req_url)

        decision_tree = self._decode_response(resp)

        return decision_tree

    def get_generator_decision_tree(
        self, generator_id, timestamp=None, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Get generator decision tree.

        :param str generator_id: the id of the generator whose tree to get. It
        must be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.
        :param int timestamp: Optional. The decision tree is comptuted
        at this timestamp.
        :default timestamp: None, means that we get the tree computed
        with all its context history.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: decision tree.
        :rtype: dict.

        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
        the tree in the time given by the configuration.
        """
        if isinstance(version, int):
            version = str(version)

        # Convert datetime to timestamp
        if isinstance(timestamp, datetime.datetime):
            timestamp = time.mktime(timestamp.timetuple())

        # Raises an error when generator_id is invalid
        self._check_entity_id(generator_id)
        if self._config["decisionTreeRetrievalTimeout"] is False:
            # Don't retry
            return self._get_generator_decision_tree(generator_id, timestamp, version)

        start = current_time_ms()
        while True:
            now = current_time_ms()
            if now - start > self._config["decisionTreeRetrievalTimeout"]:
                # Client side timeout
                raise CraftAiLongRequestTimeOutError()
            try:
                return self._get_generator_decision_tree(
                    generator_id, timestamp, version
                )
            except CraftAiLongRequestTimeOutError:
                # Do nothing and continue.
                continue

    def _get_generators_decision_trees_bulk(
        self, payload, valid_indices, invalid_indices, invalid_dts
    ):
        """Tool for the function get_generators_decision_trees_bulk.

        :param list payload: contains the informations necessary for getting
        the trees. Its form is the same than for the function.
        _get_generators_decision_trees_bulk.
        :param list valid_indices: list of the indices of the valid generator id.
        :param list invalid_indices: list of the indices of the valid generator id.
        :param list invalid_dts: list of the invalid generator id.

        :return: decision trees.
        :rtype: list of dict.
        """
        valid_dts = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/generators/tree".format(self._base_url),
            "POST",
        )

        if invalid_indices == []:
            return valid_dts

        # Put the valid and invalid decision trees in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_dts, invalid_indices, invalid_dts
        )

    def get_generators_decision_trees_bulk(
        self, payload, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Get a group of decision trees.

        :param list payload: contains the informations necessary for getting
        the trees. It's in the form [{"id": generator_id, "timestamp": timestamp}]
        With id a str containing only characters in "a-zA-Z0-9_-" and must be
        between 1 and 36 characters. It must reference an existing generator.
        With timestamp an positive and not null integer.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: Decision trees.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if all of the ids are invalid or
        referenced non existing generators or all of the timestamp are invalid.
        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
        the tree in the time given by the configuration.
        """
        if isinstance(version, int):
            version = str(version)
        self._requests_session.headers["x-craft-ai-tree-version"] = version

        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_dts = self._check_entity_id_bulk(
            payload
        )

        if self._config["decisionTreeRetrievalTimeout"] is False:
            # Don't retry
            return self._get_generators_decision_trees_bulk(
                payload, valid_indices, invalid_indices, invalid_dts
            )
        start = current_time_ms()
        while True:
            now = current_time_ms()
            if now - start > self._config["decisionTreeRetrievalTimeout"]:
                # Client side timeout
                raise CraftAiLongRequestTimeOutError()
            try:
                return self._get_generators_decision_trees_bulk(
                    payload, valid_indices, invalid_indices, invalid_dts
                )
            except CraftAiLongRequestTimeOutError:
                # Do nothing and continue.
                continue

    def _get_generator_operations_pages(self, url, ops_list):
        if url is None:
            return ops_list

        resp = self._requests_session.get(url)

        new_ops_list = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_generator_operations_pages(
            next_page_url, ops_list + new_ops_list
        )

    def get_generator_operations(self, generator_id, start=None, end=None):
        # Raises an error when generator_id is invalid
        self._check_entity_id(generator_id)

        req_url = "{}/generators/{}/context".format(self._base_url, generator_id)
        req_params = {"start": start, "end": end}
        resp = self._requests_session.get(req_url, params=req_params)

        initial_ops_list = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_generator_operations_pages(next_page_url, initial_ops_list)

    ###################
    # Context methods #
    ###################
    def add_agent_operations(self, agent_id, operations):
        """Add operations to an agent.

        :param str agent_id: The id of the agent to delete. It must be
        an str containing only characters in "a-zA-Z0-9_-" and must be
        between 1 and 36 characters. It must reference an existing agent.
        :param list operations: Contains dictionnaries that has the
        form given in the craft_ai documentation and the configuration
        of the agent.

        :return: message about the added operations.
        :rtype: str

        :raise CraftAiBadRequestError: if the input is not of
        the right form.
        """
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        # Extra header in addition to the main session's
        ct_header = {"Content-Type": "application/json; charset=utf-8"}
        offset = 0
        added_operations_count = 0

        is_looping = True

        while is_looping:
            next_offset = offset + self.config["operationsChunksSize"]

            try:
                json_pl = json.dumps(operations[offset:next_offset])
            except TypeError as err:
                raise CraftAiBadRequestError(
                    "Invalid configuration or agent id given. {}".format(err.__str__())
                )

            req_url = "{}/agents/{}/context".format(self._base_url, agent_id)
            resp = self._requests_session.post(req_url, headers=ct_header, data=json_pl)
            decoded_response = self._decode_response(resp)

            added_operations_count += extract_operations_count_from_message(
                decoded_response["message"]
            )

            if next_offset >= len(operations):
                is_looping = False

            offset = next_offset

        return {
            "message": f'Successfully added {added_operations_count} operation(s) to \
                the agent "{self.config["owner"]}/{self.config["project"]}/{agent_id}" context.',
            "added_operations_count": added_operations_count,
        }

    def _add_agents_operations_bulk(self, chunked_data):
        """Tool for the function add_agents_operations_bulk. It send the requests to
        add the operations to the agents.

        :param list chunked_data: list of list of the agents and their operations
        to add. Each chunk can be requested at the same time.

        :return: list of agents containing a message about the added
        operations.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if the input is not of the right form.
        """
        url = "{}/bulk/context".format(self._base_url)
        ct_header = {"Content-Type": "application/json; charset=utf-8"}

        responses = []
        for chunk in chunked_data:
            if len(chunk) > 1:
                try:
                    json_pl = json.dumps(chunk)
                except TypeError as err:
                    raise CraftAiBadRequestError(
                        "Error while dumping the payload into json"
                        "format when converting it for the bulk request. {}".format(
                            err.__str__()
                        )
                    )
                resp = self._requests_session.post(url, headers=ct_header, data=json_pl)
                decoded_response = self._decode_response(resp)
                responses += [
                    {
                        **r,
                        "added_operations_count": extract_operations_count_from_message(
                            r["message"]
                        ),
                    }
                    for r in decoded_response
                ]
            elif chunk:
                add_agent_operations_response = self.add_agent_operations(
                    chunk[0]["id"], chunk[0]["operations"]
                )
                responses.append(
                    {
                        "id": chunk[0]["id"],
                        "status": 201,
                        **add_agent_operations_response,
                    }
                )

        if responses == []:
            raise CraftAiBadRequestError("Invalid or empty set of operations given")

        return responses

    def add_agents_operations_bulk(self, payload):
        """Add operations to a group of agents.

        :param list payload: contains the informations necessary for the action.
        It's in the form [{"id": agent_id, "operations": operations}]
        With id that is an str containing only characters in "a-zA-Z0-9_-"
        and must be between 1 and 36 characters. It must reference an
        existing agent.
        With operations a list containing dictionnaries that has the form given
        in the craft_ai documentation and the configuration of the agent.

        :return: list of agents containing a message about the added
        operations.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if all of the ids are invalid or
        referenced non existing agents or one of the operations is invalid.
        """
        # Check all ids, raise an error if all ids are invalid
        valid_indices, _, _ = self._check_entity_id_bulk(payload)
        valid_payload = [payload[i] for i in valid_indices]

        chunked_data = []
        current_chunk = []
        current_chunk_size = 0

        for agent in valid_payload:
            if agent["operations"] and isinstance(agent["operations"], list):
                if (
                    current_chunk_size + len(agent["operations"])
                    > self.config["operationsChunksSize"]
                ):
                    chunked_data.append(current_chunk)
                    current_chunk_size = 0
                    current_chunk = []
                if len(agent["operations"]) > self.config["operationsChunksSize"]:
                    chunked_data.append([agent])
                    current_chunk_size = 0
                else:
                    current_chunk_size += len(agent["operations"])
                    current_chunk.append(agent)

        if current_chunk:
            chunked_data.append(current_chunk)

        return self._add_agents_operations_bulk(chunked_data)

    def _get_agent_operations_pages(self, url, ops_list):
        if url is None:
            return ops_list

        resp = self._requests_session.get(url)

        new_ops_list = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_agent_operations_pages(next_page_url, ops_list + new_ops_list)

    def get_agent_operations(self, agent_id, start=None, end=None):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}/context".format(self._base_url, agent_id)
        req_params = {"start": start, "end": end}
        resp = self._requests_session.get(req_url, params=req_params)

        initial_ops_list = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_agent_operations_pages(next_page_url, initial_ops_list)

    def _get_agent_states_pages(self, url, states):
        if url is None:
            return states

        resp = self._requests_session.get(url)

        new_states = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_agent_states_pages(next_page_url, states + new_states)

    def get_agent_states(self, agent_id, start=None, end=None):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}/context/state/history".format(self._base_url, agent_id)
        req_params = {"start": start, "end": end}
        resp = self._requests_session.get(req_url, params=req_params)

        initial_states_history = self._decode_response(resp)
        next_page_url = resp.headers.get("x-craft-ai-next-page-url")

        return self._get_agent_states_pages(next_page_url, initial_states_history)

    def get_agent_state(self, agent_id, timestamp):
        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)

        req_url = "{}/agents/{}/context/state?t={}".format(
            self._base_url, agent_id, timestamp
        )
        resp = self._requests_session.get(req_url)

        context_state = self._decode_response(resp)

        return context_state

    #########################
    # Decision tree methods #
    #########################

    def _get_agent_decision_tree(
        self, agent_id, timestamp, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Tool for the function get_agent_decision_tree.

        :param str agent_id: the id of the agent whose tree to get. It
        must be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.
        :param int timestamp: Optional. The decision tree is comptuted
        at this timestamp.
        :default timestamp: None, means that we get the tree computed
        with all its context history.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: decision tree.
        :rtype: dict.
        """
        self._requests_session.headers["x-craft-ai-tree-version"] = version

        # If we give no timestamp the default behaviour is to give
        # the tree from the latest timestamp
        if timestamp is None:
            req_url = "{}/agents/{}/decision/tree?".format(self._base_url, agent_id)
        else:
            req_url = "{}/agents/{}/decision/tree?t={}".format(
                self._base_url, agent_id, timestamp
            )

        resp = self._requests_session.get(req_url)

        decision_tree = self._decode_response(resp)

        return decision_tree

    def get_agent_decision_tree(
        self, agent_id, timestamp=None, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Get decision tree.

        :param str agent_id: the id of the agent whose tree to get. It
        must be an str containing only characters in "a-zA-Z0-9_-" and
        must be between 1 and 36 characters.
        :param int timestamp: Optional. The decision tree is comptuted
        at this timestamp.
        :default timestamp: None, means that we get the tree computed
        with all its context history.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: decision tree.
        :rtype: dict.

        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
        the tree in the time given by the configuration.
        """
        if isinstance(version, int):
            version = str(version)

        # Convert datetime to timestamp
        if isinstance(timestamp, datetime.datetime):
            timestamp = time.mktime(timestamp.timetuple())

        # Raises an error when agent_id is invalid
        self._check_entity_id(agent_id)
        if self._config["decisionTreeRetrievalTimeout"] is False:
            # Don't retry
            return self._get_agent_decision_tree(agent_id, timestamp, version)

        start = current_time_ms()
        while True:
            now = current_time_ms()
            if now - start > self._config["decisionTreeRetrievalTimeout"]:
                # Client side timeout
                raise CraftAiLongRequestTimeOutError()
            try:
                return self._get_agent_decision_tree(agent_id, timestamp, version)
            except CraftAiLongRequestTimeOutError:
                # Do nothing and continue.
                continue

    def _get_agents_decision_trees_bulk(
        self, payload, valid_indices, invalid_indices, invalid_dts
    ):
        """Tool for the function get_agents_decision_trees_bulk.

        :param list payload: contains the informations necessary for getting
        the trees. Its form is the same than for the function.
        get_agents_decision_trees_bulk.
        :param list valid_indices: list of the indices of the valid agent id.
        :param list invalid_indices: list of the indices of the valid agent id.
        :param list invalid_dts: list of the invalid agent id.

        :return: decision trees.
        :rtype: list of dict.
        """
        valid_dts = self._create_and_send_json_bulk(
            [payload[i] for i in valid_indices],
            "{}/bulk/decision_tree".format(self._base_url),
            "POST",
        )

        if invalid_indices == []:
            return valid_dts

        # Put the valid and invalid decision trees in their original index
        return self._recreate_list_with_indices(
            valid_indices, valid_dts, invalid_indices, invalid_dts
        )

    def get_agents_decision_trees_bulk(
        self, payload, version=DEFAULT_DECISION_TREE_VERSION
    ):
        """Get a group of decision trees.

        :param list payload: contains the informations necessary for getting
        the trees. It's in the form [{"id": agent_id, "timestamp": timestamp}]
        With id a str containing only characters in "a-zA-Z0-9_-" and must be
        between 1 and 36 characters. It must reference an existing agent.
        With timestamp an positive and not null integer.
        :param version: version of the tree to get.
        :type version: str or int.
        :default version: default version of the tree.

        :return: Decision trees.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if all of the ids are invalid or
        referenced non existing agents or all of the timestamp are invalid.
        :raises CraftAiLongRequestTimeOutError: if the API doesn't get
        the tree in the time given by the configuration.
        """
        if isinstance(version, int):
            version = str(version)
        self._requests_session.headers["x-craft-ai-tree-version"] = version

        # Check all ids, raise an error if all ids are invalid
        valid_indices, invalid_indices, invalid_dts = self._check_entity_id_bulk(
            payload
        )

        if self._config["decisionTreeRetrievalTimeout"] is False:
            # Don't retry
            return self._get_agents_decision_trees_bulk(
                payload, valid_indices, invalid_indices, invalid_dts
            )
        start = current_time_ms()
        while True:
            now = current_time_ms()
            if now - start > self._config["decisionTreeRetrievalTimeout"]:
                # Client side timeout
                raise CraftAiLongRequestTimeOutError()
            try:
                return self._get_agents_decision_trees_bulk(
                    payload, valid_indices, invalid_indices, invalid_dts
                )
            except CraftAiLongRequestTimeOutError:
                # Do nothing and continue.
                continue

    @staticmethod
    def decide(tree, *args):
        return Interpreter.decide(tree, args)

    @staticmethod
    def _parse_body(response):
        try:
            return response.json()
        except Exception:
            raise CraftAiInternalError(
                "Internal Error, the craft ai server responded in an invalid format."
            )

    @staticmethod
    def _decode_response(response):
        """Decode the response of a request.

        :param response: response of a request.

        :return: decoded response.

        :raise Error: Raise the error given by the request.
        """
        status_code = response.status_code

        message = "Status code " + str(status_code)
        try:
            message = Client._parse_body(response)["message"]
        except (CraftAiInternalError, KeyError, TypeError):
            pass

        if status_code in [200, 201, 204, 207]:
            return Client._parse_body(response)
        else:
            raise Client._get_error_from_status(status_code, message)
        return None

    @staticmethod
    def _decode_response_bulk(response_bulk):
        """Decode the response of each agent given by a bulk function.

        :param list response_bulk: list of dictionnary which represents
        the response for an agent.

        :return: decoded response.
        :rtype: list of dict.
        """
        resp = []
        for response in response_bulk:
            if ("status" in response) and (response.get("status") == 201):
                agent = {"id": response["id"], "message": response["message"]}
                resp.append(agent)
            elif "status" in response:
                status_code = response["status"]
                message = response["message"]
                agent = {"error": Client._get_error_from_status(status_code, message)}
                try:
                    agent["id"] = response["id"]
                except KeyError:
                    pass
                resp.append(agent)

            else:
                resp.append(response)

        return resp

    @staticmethod
    def _get_error_from_status(status_code, message):
        """Give the error corresponding to the status code.

        :param int status_code: status code of the response to
        a request.
        :param str message: error message given by the response.

        :return: error corresponding to the status code.
        :rtype: Error.
        """
        if status_code == 202:
            err = CraftAiLongRequestTimeOutError(message)
        elif status_code == 401 or status_code == 403:
            err = CraftAiCredentialsError(message)
        elif status_code == 400:
            err = CraftAiBadRequestError(message)
        elif status_code == 404:
            err = CraftAiNotFoundError(message)
        elif status_code == 413:
            err = CraftAiBadRequestError("Given payload is too large")
        elif status_code == 500:
            err = CraftAiInternalError(message)
        elif status_code == 503:
            err = CraftAiNetworkError(
                """Service momentarily unavailable, please try """
                """again in a few minutes. If the problem """
                """persists please contact us at support@craft.ai"""
            )
        elif status_code == 504:
            err = CraftAiBadRequestError("Request has timed out")
        else:
            err = CraftAiUnknownError(message)

        return err

    @staticmethod
    def _check_entity_id(entity_id):
        """Checks that the given entity_id is a valid non-empty string.

        :param str entity_id: entity id to check.

        :raise CraftAiBadRequestError: If the given entity_id is not of
        type string or if it is an empty string.
        """
        if not isinstance(entity_id, str) or AGENT_ID_PATTERN.match(entity_id) is None:
            raise CraftAiBadRequestError(ERROR_ID_MESSAGE)

    def _check_entity_id_bulk(self, payload, check_serializable=True):
        """Checks that all the given entity ids are valid non-empty strings
        and if the entities are serializable.

        :param list payload: list of dictionnary which represents an entity.

        :return: list of the entities with valid ids, list of the entities with
        invalid ids, list of the dictionnaries with valid ids.
        :rtype: list, list, list of dict.

        :raise CraftAiBadRequestError: If all the entities are invalid.
        """
        invalid_entity_indices = []
        valid_entity_indices = []
        invalid_payload = []
        for index, entity in enumerate(payload):
            # Check if the entity ID is valid
            try:
                if "id" in entity:
                    self._check_entity_id(entity["id"])
            except CraftAiBadRequestError:
                invalid_entity_indices.append(index)
                invalid_payload.append(
                    {
                        "id": entity["id"],
                        "error": CraftAiBadRequestError(ERROR_ID_MESSAGE),
                    }
                )
            else:
                if check_serializable:
                    # Check if the entity is serializable
                    try:
                        json.dumps([entity])
                    except TypeError as err:
                        invalid_entity_indices.append(index)
                        invalid_payload.append({"id": entity["id"], "error": err})
                    else:
                        valid_entity_indices.append(index)
                else:
                    valid_entity_indices.append(index)

        if len(invalid_entity_indices) == len(payload):
            raise CraftAiBadRequestError(ERROR_ID_MESSAGE)

        return valid_entity_indices, invalid_entity_indices, invalid_payload

    @staticmethod
    def _recreate_list_with_indices(indices1, values1, indices2, values2):
        """Create a list in the right order.

        :param list indices1: contains the list of indices corresponding to
        the values in values1.
        :param list values1: contains the first list of values.
        :param list indices2: contains the list of indices corresponding to
        the values in values2.
        :param list values2: contains the second list of values.

        :return: list of the values in the correct order.
        :rtype: list.
        """
        # Test if indices are continuous
        list_indices = sorted(indices1 + indices2)
        for i, index in enumerate(list_indices):
            if i != index:
                raise CraftAiInternalError("The agents's indices are not continuous")

        full_list = [None] * (len(indices1) + len(indices2))
        for i, index in enumerate(indices1):
            full_list[index] = values1[i]
        for i, index in enumerate(indices2):
            full_list[index] = values2[i]
        return full_list

    def _create_and_send_json_bulk(self, payload, req_url, request_type="POST"):
        """Create a json, do a request to the URL and process the response.

        :param list payload: contains the informations necessary for the action.
        It's a list of dictionnary.
        :param str req_url: URL to request with the payload.
        :param str request_type: type of request, either "POST" or "DELETE".
        :default request_type: "POST".

        :return: response of the request.
        :rtype: list of dict.

        :raises CraftAiBadRequestError: if the payload doesn't have the
        correct form to be transformed into a json or request_type is
        neither "POST" or "DELETE".
        """
        # Extra header in addition to the main session's
        ct_header = {"Content-Type": "application/json; charset=utf-8"}

        try:
            json_pl = json.dumps(payload)
        except TypeError as err:
            raise CraftAiBadRequestError(
                "Error while dumping the payload into json"
                "format when converting it for the bulk request. {}".format(
                    err.__str__()
                )
            )
        if request_type == "POST":
            resp = self._requests_session.post(req_url, headers=ct_header, data=json_pl)
        elif request_type == "DELETE":
            resp = self._requests_session.delete(
                req_url, headers=ct_header, data=json_pl
            )
        else:
            raise CraftAiBadRequestError(
                "Request for the bulk API should be either a POST or DELETE" "request"
            )
        entities = self._decode_response(resp)
        entities = self._decode_response_bulk(entities)
        return entities



class MissingValue(object):
    def __str__(self):
        return "MISSING"


class OptionalValue(object):
    def __str__(self):
        return "OPTIONAL"


MISSING_VALUE = MissingValue()
OPTIONAL_VALUE = OptionalValue()
